import collections
import functools

from django.db import models
from django.db.models import fields, options

# Django doesn't support adding third-party fields to model class Meta.
if 'repr_fields' not in options.DEFAULT_NAMES:
    options.DEFAULT_NAMES = options.DEFAULT_NAMES + (
        'repr_fields', 'str_fields')

_model_name_counter = collections.Counter()
_dunder_applied_counter = 0

FIELD_REPR_FMT="{}={!r}"
FIELD_STR_FMT="{}={}"


def field_value(model, field):
    value = getattr(model, field.name)

    default = field.default
    if value == default:
        return None
    elif default is models.NOT_PROVIDED:
        if isinstance(field, models.CharField) and value == '':
            return None

    return value


def _format_field(model, field, fmt=FIELD_REPR_FMT):
    """Obtain either "bar_id=9" or "bar='bar_value'" str.

    field is a Django Field object.
    """
    if isinstance(field, models.ForeignKey):
        field_name = field.name + '_id'
    else:
        field_name = field.name
    return _format_field_raw(model, field, field_name, fmt=fmt)


def _format_field_raw(model, field, field_name=None, fmt=FIELD_REPR_FMT):
    """Obtain "bar='bar_value'" str.

    field is a Django Field object.
    field_name is either field.name or field.name + '_id'
    """
    value = field_value(model, field)
    if not value:
        return ""

    return fmt.format(field_name or field.name, value)


def _to_string(model, meta_field_name, fmt):
    self = model
    cls = type(self)
    meta_fields = cls._meta.fields
    has_autofield = isinstance(meta_fields[0], fields.AutoField)
    selected_field_names = None
    if hasattr(cls._meta, meta_field_name):
        selected_field_names = getattr(cls._meta, meta_field_name)
    elif cls._meta.unique_together:
        selected_field_names = cls._meta.unique_together[0]
    elif len(meta_fields) > 1:
        uniq_fields = [f for f in (meta_fields[1:] if has_autofield else meta_fields) if f.unique]
        if uniq_fields:
            # TODO: determine which is best
            selected_field_names = [uniq_fields[0].name]

    if selected_field_names:
        meta_fields_filtered = [f for f in meta_fields if f.name in selected_field_names]
        meta_fields_filtered.sort(key=lambda f: selected_field_names.index(f.name))

        func = functools.partial(_format_field_raw, self, fmt=fmt)
        parts = filter(None, map(func, meta_fields_filtered))
    elif len(meta_fields) < 3:
        # _raw shows 'other=Foo' instead of other_id=3
        func = functools.partial(_format_field_raw, self, fmt=fmt)
        parts = filter(None, map(func, meta_fields))
    else:
        func = functools.partial(_format_field, self, fmt=fmt)
        parts = filter(None, map(func, meta_fields))

    parts = list(parts)

    if has_autofield and len(meta_fields) == 2 and len(parts) == 2:
        # Ignore unnecessary 'id'
        parts = [parts[1]]

    attrs = ', '.join(parts)

    if not attrs and has_autofield:
        attrs = _format_field_raw(self, meta_fields[0], fmt=fmt)

    if _model_name_counter[cls.__name__] > 1:
        model_name = cls._meta.label
    else:
        model_name = cls.__name__

    rv = '{}({})'.format(model_name, attrs)
    return rv


def _model_repr(self):
    return _to_string(
        model=self,
        meta_field_name='repr_fields',
        fmt=FIELD_REPR_FMT,
    )


def _model_str(self):
    return _to_string(
        model=self,
        meta_field_name='str_fields',
        fmt=FIELD_STR_FMT,
    )



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _

from .app_settings import _ANY_REGISTER


class DunderConfig(AppConfig):
    """The default dunder configuration."""

    label = 'dunder'
    name = 'django_dunder'
    verbose_name = _('Django dunders')

    def ready(self):
        # The registration should already have been done
        if _ANY_REGISTER:
            import django_dunder._register  # noqa

    def get_models(self, *args, **kwargs):
        # The following is used to prevent the .models from being loaded
        # especially by test harnesses
        return []



import sys

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

def get_setting_safe(name, default):
    try:
        return getattr(settings, 'DUNDER_' + name, default)
    except ImproperlyConfigured:
        return default

AUTO = get_setting_safe('AUTO', True)

AUTO_REPR = get_setting_safe('AUTO_REPR', AUTO)
AUTO_STR = get_setting_safe('AUTO_STR', AUTO)

FORCE = get_setting_safe('FORCE', False)

FORCE_REPR = get_setting_safe('FORCE_REPR', False)
FORCE_STR = get_setting_safe('FORCE_STR', False)

_ANY_REGISTER = (
    AUTO or AUTO_REPR or AUTO_STR or FORCE or FORCE_REPR or FORCE_STR)

REPR_EXCLUDE = get_setting_safe('REPR_EXCLUDE', False)
STR_EXCLUDE = get_setting_safe('STR_EXCLUDE', False)



from django.db.models import Model
from django.db.models.signals import class_prepared
from django.dispatch import receiver

from .app_settings import (
    _ANY_REGISTER,
    AUTO,
    AUTO_REPR,
    AUTO_STR,
    FORCE,
    FORCE_REPR,
    FORCE_STR,
    REPR_EXCLUDE,
    STR_EXCLUDE,
)
from .core import (
    _dunder_applied_counter,
    _model_name_counter,
    _model_repr,
    _model_str,
)


def _has_default_repr(model):
    for cls in model.__mro__:
        if '__repr__' in cls.__dict__:
            if cls.__repr__ == Model.__repr__:
                return True
            else:
                return False


def _has_default_str(model):
    for cls in model.__mro__:
        if '__str__' in cls.__dict__:
            if cls.__str__ == Model.__str__:
                return True
            else:
                return False


def _should_force_repr(label):
    if FORCE_REPR is False:
        return False

    if isinstance(FORCE_REPR, list):
        return label in FORCE_REPR

    if REPR_EXCLUDE:
        return label not in REPR_EXCLUDE

    return FORCE

def _should_force_str(label):
    if FORCE_STR is False:
        return False

    if isinstance(FORCE_STR, list):
        return label in FORCE_STR

    if STR_EXCLUDE:
        return label not in STR_EXCLUDE

    return FORCE


@receiver(class_prepared)
def _patch_model_cls(sender, **kwargs):
    global _dunder_applied_counter, _model_name_counter

    # This is used to prefix duplicated names
    _model_name_counter[sender.__class__.__name__] += 1

    if not _ANY_REGISTER:
        return

    label = sender._meta.label

    if sender.__repr__ != _model_repr:
        if _should_force_repr(sender) or (AUTO_REPR and _has_default_repr(sender)):
            if not REPR_EXCLUDE or label not in REPR_EXCLUDE:
                _dunder_applied_counter += 1
                sender.__repr__ = _model_repr

    if sender.__str__ != _model_str:
        if _should_force_str(label) or (AUTO_STR and _has_default_str(sender)):
            if not STR_EXCLUDE or label not in STR_EXCLUDE:
                _dunder_applied_counter += 1
                sender.__str__ = _model_str



from django.apps import apps
from django.conf import settings
from django.db import models

from django_nine.versions import DJANGO_GTE_2_0

from django_dunder.core import _model_name_counter
from django_dunder.mixins import DunderStrModel

from django_fake_model import models as f


class StrDefaultOrig(f.FakeModel):
    name1 = models.TextField(null=True, blank=True)
    name2 = models.TextField(null=True, blank=True)


class StrDefault(DunderStrModel, f.FakeModel):
    name1 = models.TextField(null=True, blank=True)
    name2 = models.TextField(null=True, blank=True)


@StrDefaultOrig.fake_me
@StrDefault.fake_me
def test_str_default():
    item_orig = StrDefaultOrig.objects.create(name1='a', name2='b')

    if DJANGO_GTE_2_0:
        assert str(item_orig) == "StrDefaultOrig object (1)"
    else:
        assert str(item_orig) == "StrDefaultOrig object"

    item = StrDefault.objects.create(name1='a', name2='b')

    assert str(item) == "StrDefault(id=1, name1=a, name2=b)"


class StrUniqueOrig(f.FakeModel):
    name1 = models.TextField(unique=True)
    name2 = models.TextField(null=True, blank=True)


class StrUnique(DunderStrModel, f.FakeModel):
    name1 = models.TextField(unique=True)
    name2 = models.TextField(null=True, blank=True)


@StrUniqueOrig.fake_me
@StrUnique.fake_me
def test_str_unique():
    item_orig = StrUniqueOrig.objects.create(name1='a', name2='b')

    if DJANGO_GTE_2_0:
        assert str(item_orig) == "StrUniqueOrig object (1)"
    else:
        assert str(item_orig) == "StrUniqueOrig object"

    item = StrUnique.objects.create(name1='a', name2='b')

    assert str(item) == "StrUnique(name1=a)"


class StrPrimaryKeyOrig(f.FakeModel):
    name1 = models.TextField(primary_key=True)
    name2 = models.TextField(null=True, blank=True)


class StrPrimaryKey(DunderStrModel, f.FakeModel):
    name1 = models.TextField(primary_key=True)
    name2 = models.TextField(null=True, blank=True)


@StrPrimaryKeyOrig.fake_me
@StrPrimaryKey.fake_me
def test_str_primary_key():
    item_orig = StrPrimaryKeyOrig.objects.create(name1='a', name2='b')

    if DJANGO_GTE_2_0:
        assert str(item_orig) == "StrPrimaryKeyOrig object (a)"
    else:
        assert str(item_orig) == "StrPrimaryKeyOrig object"

    item = StrPrimaryKey.objects.create(name1='a', name2='b')

    assert str(item) == "StrPrimaryKey(name1=a)"


class StrExplicit(DunderStrModel, f.FakeModel):
    name1 = models.TextField(null=True, blank=True)
    name2 = models.TextField(null=True, blank=True)

    class Meta:
        str_fields = ('name2', 'name1')


@StrExplicit.fake_me
def test_str_explicit():
    item = StrExplicit.objects.create(name1='a', name2='b')

    assert str(item) == "StrExplicit(name2=b, name1=a)"

    item = StrExplicit.objects.create(name2='b')

    assert str(item) == "StrExplicit(name2=b)"

    item = StrExplicit.objects.create(name1='a')

    assert str(item) == "StrExplicit(name1=a)"


class StrUniqueTogether(DunderStrModel, f.FakeModel):
    name = models.TextField(null=True, blank=True)
    name1 = models.TextField(null=True, blank=True)
    name2 = models.TextField(null=True, blank=True)

    class Meta:
        unique_together = ('name1', 'name2')


@StrUniqueTogether.fake_me
def test_str_unique_together():
    item = StrUniqueTogether.objects.create(
        name="z", name1="a", name2="b")

    assert str(item) == "StrUniqueTogether(name1=a, name2=b)"


class StrOneField(DunderStrModel, f.FakeModel):
    name = models.TextField(null=True, blank=True)


@StrOneField.fake_me
def test_str_one_field():
    item = StrOneField.objects.create()

    assert str(item) == 'StrOneField(id=1)'

    item = StrOneField.objects.create(name='a')

    assert str(item) == "StrOneField(name=a)"


class StrNameConflict1(DunderStrModel, f.FakeModel):

    class Meta:
        app_label = 'myapp1'


class StrNameConflict2(DunderStrModel, f.FakeModel):

    class Meta:
        app_label = 'myapp2'


@StrNameConflict1.fake_me
@StrNameConflict2.fake_me
def test_str_explicit_app():
    # Simulate them having the same name in different apps
    _model_name_counter['StrNameConflict1'] = 2
    _model_name_counter['StrNameConflict2'] = 2

    item1 = StrNameConflict1.objects.create()
    item2 = StrNameConflict2.objects.create()

    assert str(item1) == 'myapp1.StrNameConflict1(id=1)'
    assert str(item2) == 'myapp2.StrNameConflict2(id=1)'



from .app_settings import _ANY_REGISTER

# Start registration of models appearing after this app in INSTALLED_APPS
if _ANY_REGISTER:
    from ._register import _dunder_applied_counter
else:
    _dunder_applied_counter = 0

__all__ = [
    '_dunder_applied_counter',
]




from django.db import models

from .core import _model_repr, _model_str

__all__ = [
    'DunderReprModel',
    'DunderStrModel',
    'DunderModel',
]


class DunderReprModel(models.Model):
    class Meta:
        abstract = True

    __repr__ = _model_repr


class DunderStrModel(models.Model):
    class Meta:
        abstract = True

    __str__ = _model_str


class DunderModel(DunderReprModel, DunderStrModel):
    class Meta:
        abstract = True



import io
import os
import re
import sys

from setuptools import setup, find_packages

PATH_BASE = os.path.dirname(__file__)
PACKAGE_NAME = 'django-dunder'
PACKAGE_DIR = 'django_dunder'

def read_file(fpath):
    """Reads a file within package directories."""
    with io.open(os.path.join(PATH_BASE, fpath)) as f:
        return f.read()


def get_version():
    """Returns version number, without module import (which can lead to
    ImportError if some dependencies are unavailable before install."""
    contents = read_file(os.path.join(PACKAGE_DIR, '__init__.py'))
    version = re.search('VERSION = \(([^)]+)\)', contents)
    version = version.group(1).replace(', ', '.').strip()
    return version


setup(
    name=PACKAGE_NAME,
    version=get_version(),
    url='http://github.com/jayvdb/' + PACKAGE_NAME,

    description='Reusable Django app to automate dunders `__repr__` and `__str__`',
    long_description=read_file('README.md'),
    long_description_content_type='text/markdown',
    license='MIT',

    author='John Mark Vandenberg',
    author_email='jayvdb@gmail.com',

    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,

    test_suite=PACKAGE_DIR + '.tests',
    tests_require=[
        'pytest-djangoapp',
        'django-fake-model',
        'django-nine',
    ],

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)



from django.db import models

from django_nine.versions import DJANGO_GTE_2_0

from django_dunder.core import _model_name_counter
from django_dunder.mixins import DunderReprModel

from django_fake_model import models as f


class ReprDefaultOrig(f.FakeModel):
    name1 = models.TextField(null=True, blank=True)
    name2 = models.TextField(null=True, blank=True)


class ReprDefault(DunderReprModel, f.FakeModel):
    name1 = models.TextField(null=True, blank=True)
    name2 = models.TextField(null=True, blank=True)


@ReprDefaultOrig.fake_me
@ReprDefault.fake_me
def test_repr_default():
    item_orig = ReprDefaultOrig.objects.create(name1='a', name2='b')

    if DJANGO_GTE_2_0:
        assert repr(item_orig) == "<ReprDefaultOrig: ReprDefaultOrig object (1)>"
    else:
        assert repr(item_orig) == "<ReprDefaultOrig: ReprDefaultOrig object>"

    item = ReprDefault.objects.create(name1='a', name2='b')

    assert repr(item) == "ReprDefault(id=1, name1='a', name2='b')"


class Explicit(DunderReprModel, f.FakeModel):
    name1 = models.TextField(null=True, blank=True)
    name2 = models.TextField(null=True, blank=True)

    class Meta:
        repr_fields = ('name2', 'name1')


@Explicit.fake_me
def test_repr_explicit():
    item = Explicit.objects.create(name1='a', name2='b')

    assert repr(item) == "Explicit(name2='b', name1='a')"

    item = Explicit.objects.create(name2='b')

    assert repr(item) == "Explicit(name2='b')"

    item = Explicit.objects.create(name1='a')

    assert repr(item) == "Explicit(name1='a')"


class UniqueTogether(DunderReprModel, f.FakeModel):
    name = models.TextField(null=True, blank=True)
    name1 = models.TextField(null=True, blank=True)
    name2 = models.TextField(null=True, blank=True)

    class Meta:
        unique_together = ('name1', 'name2')


@UniqueTogether.fake_me
def test_repr_unique_together():
    item = UniqueTogether.objects.create(
        name="z", name1="a", name2="b")

    assert repr(item) == "UniqueTogether(name1='a', name2='b')"


class OneField(DunderReprModel, f.FakeModel):
    name = models.TextField(null=True, blank=True)


@OneField.fake_me
def test_repr_one_field():
    item = OneField.objects.create()

    assert repr(item) == 'OneField(id=1)'

    item = OneField.objects.create(name='a')

    assert repr(item) == "OneField(name='a')"


class NameConflict1(DunderReprModel, f.FakeModel):

    class Meta:
        app_label = 'myapp1'


class NameConflict2(DunderReprModel, f.FakeModel):

    class Meta:
        app_label = 'myapp2'


@NameConflict1.fake_me
@NameConflict2.fake_me
def test_repr_explicit_app():
    # Simulate them having the same name in different apps
    _model_name_counter['NameConflict1'] = 2
    _model_name_counter['NameConflict2'] = 2

    item1 = NameConflict1.objects.create()
    item2 = NameConflict2.objects.create()

    assert repr(item1) == 'myapp1.NameConflict1(id=1)'
    assert repr(item2) == 'myapp2.NameConflict2(id=1)'
